package com.safeway.epe.avrokafkawithoutschemaregistry.producer;

import io.confluent.kafka.schemaregistry.RestApp;
import io.confluent.kafka.schemaregistry.avro.AvroCompatibilityLevel;
import io.confluent.kafka.schemaregistry.rest.SchemaRegistryConfig;
import kafka.server.KafkaConfig;
import kafka.server.KafkaConfig$;
import kafka.server.KafkaServer;
import org.apache.curator.test.TestingServer;
import org.apache.kafka.clients.admin.AdminClient;
import org.apache.kafka.clients.admin.AdminClientConfig;
import org.apache.kafka.clients.admin.NewTopic;
import org.apache.kafka.common.network.ListenerName;
import org.apache.kafka.common.security.auth.SecurityProtocol;
import org.apache.kafka.common.utils.Time;
import org.apache.kafka.test.TestCondition;
import org.apache.kafka.test.TestUtils;
import org.junit.jupiter.api.extension.AfterEachCallback;
import org.junit.jupiter.api.extension.BeforeEachCallback;
import org.junit.jupiter.api.extension.ExtensionContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.Collections;
import java.util.Comparator;
import java.util.Map;
import java.util.Properties;
import java.util.concurrent.ExecutionException;


public class EmbeddedSingleNodeKafkaCluster implements BeforeEachCallback, AfterEachCallback {

    private static final Logger log = LoggerFactory.getLogger(EmbeddedSingleNodeKafkaCluster.class);
    private static final String KAFKA_SCHEMAS_TOPIC = "_schemas";
    private static final String AVRO_COMPATIBILITY_TYPE = AvroCompatibilityLevel.NONE.name;

    private static final String KAFKASTORE_OPERATION_TIMEOUT_MS = "60000";
    private static final String KAFKASTORE_INIT_TIMEOUT = "90000";

    private ZooKeeperEmbedded zookeeper;
    private KafkaEmbedded broker;
    private RestApp schemaRegistry;
    private final Properties brokerConfig;
    private boolean running;

    public EmbeddedSingleNodeKafkaCluster() {
        this(new Properties());
    }

    public EmbeddedSingleNodeKafkaCluster(final Properties brokerConfig) {
        this.brokerConfig = new Properties();
        this.brokerConfig.put(SchemaRegistryConfig.KAFKASTORE_TIMEOUT_CONFIG, KAFKASTORE_OPERATION_TIMEOUT_MS);
        this.brokerConfig.putAll(brokerConfig);
    }

    public void start() throws Exception {
        log.debug("Initiating embedded Kafka cluster startup");
        log.debug("Starting a ZooKeeper instance...");
        zookeeper = new ZooKeeperEmbedded();
        log.debug("ZooKeeper instance is running at {}", zookeeper.connectString());
        log.debug("ZooKeeper HostName: {}", zookeeper.hostname());
        final Properties effectiveBrokerConfig = effectiveBrokerConfigFrom(brokerConfig, zookeeper);
        broker = new KafkaEmbedded(effectiveBrokerConfig);
        log.debug("Starting a Kafka instance on port {} ...", effectiveBrokerConfig.getProperty(KafkaConfig$.MODULE$.PortProp()));
        log.info("Kafka instance is running at {}, connected to ZooKeeper at {}", bootstrapServers(), zookeeperConnect());

        final Properties schemaRegistryProps = new Properties();

        schemaRegistryProps.put(SchemaRegistryConfig.KAFKASTORE_TIMEOUT_CONFIG, KAFKASTORE_OPERATION_TIMEOUT_MS);
        schemaRegistryProps.put(SchemaRegistryConfig.KAFKASTORE_INIT_TIMEOUT_CONFIG, KAFKASTORE_INIT_TIMEOUT);
        schemaRegistryProps.put(SchemaRegistryConfig.KAFKASTORE_BOOTSTRAP_SERVERS_CONFIG, bootstrapServers());
        schemaRegistryProps.put(SchemaRegistryConfig.HOST_NAME_CONFIG, "127.0.0.1");

        schemaRegistry = new RestApp(8081, zookeeperConnect(), KAFKA_SCHEMAS_TOPIC, AVRO_COMPATIBILITY_TYPE, schemaRegistryProps);
        schemaRegistry.start();
        log.info("SchemaRegistry URL: {}", schemaRegistryUrl());
        running = true;
    }

    private Properties effectiveBrokerConfigFrom(final Properties brokerConfig, final ZooKeeperEmbedded zookeeper) {
        final Properties effectiveConfig = new Properties();
        effectiveConfig.putAll(brokerConfig);
        effectiveConfig.put(KafkaConfig$.MODULE$.ZkConnectProp(), zookeeper.connectString());
        effectiveConfig.put(KafkaConfig$.MODULE$.ZkSessionTimeoutMsProp(), 30 * 1000);
        effectiveConfig.put(KafkaConfig$.MODULE$.ZkConnectionTimeoutMsProp(), 60 * 1000);
        effectiveConfig.put(KafkaConfig$.MODULE$.DeleteTopicEnableProp(), true);
        effectiveConfig.put(KafkaConfig$.MODULE$.PortProp(), "9092");
        effectiveConfig.put(KafkaConfig$.MODULE$.LogCleanerDedupeBufferSizeProp(), 2 * 1024 * 1024L);
        effectiveConfig.put(KafkaConfig$.MODULE$.GroupMinSessionTimeoutMsProp(), 0);
        effectiveConfig.put(KafkaConfig$.MODULE$.OffsetsTopicReplicationFactorProp(), (short) 1);
        effectiveConfig.put(KafkaConfig$.MODULE$.OffsetsTopicPartitionsProp(), 1);
        effectiveConfig.put(KafkaConfig$.MODULE$.AutoCreateTopicsEnableProp(), true);
        return effectiveConfig;
    }

    @Override
    public void afterEach(ExtensionContext context) throws IOException {
        stop();
    }

    @Override
    public void beforeEach(ExtensionContext context) throws Exception {
        start();
    }

    public void stop() throws IOException {
        log.info("Stopping Confluent");
        try {
            try {
                if (schemaRegistry != null) {
                    schemaRegistry.stop();
                }
            } catch (final Exception fatal) {
                log.error(fatal.getMessage());
            }
            if (broker != null) {
                broker.stop();
            }
            try {
                if (zookeeper != null) {
                    zookeeper.stop();
                }
            } catch (final IOException fatal) {
                throw new RuntimeException(fatal);
            }
        } finally {
            if (isRunning())
                running = false;
        }
        log.info("Confluent Stopped");
    }

    public String bootstrapServers() {
        return broker.brokerList();
    }

    private String zookeeperConnect() {
        return zookeeper.connectString();
    }

    public String schemaRegistryUrl() {
        return schemaRegistry.restConnect;
    }

    public void createTopic(final String topic) throws InterruptedException {
        log.info("Topic Created: {}", topic);
        createTopic(topic, 1, (short) 1);
    }

    public void createTopic(final String topic, final int partitions, final short replication) throws InterruptedException {
        createTopic(topic, partitions, replication, Collections.emptyMap());
    }

    public void createTopic(final String topic,
                            final int partitions,
                            final short replication,
                            final Map<String, String> topicConfig) throws InterruptedException {
        createTopic(60000L, topic, partitions, replication, topicConfig);
    }

    public void createTopic(final long timeoutMs,
                            final String topic,
                            final int partitions,
                            final short replication,
                            final Map<String, String> topicConfig) throws InterruptedException {
        broker.createTopic(topic, partitions, replication, topicConfig);

        if (timeoutMs > 0) {
            TestUtils.waitForCondition(new TopicCreatedCondition(topic), timeoutMs, "Topics not created after " + timeoutMs + " milli seconds.");
        }
    }

    public boolean isRunning() {
        return running;
    }

    private final class TopicCreatedCondition implements TestCondition {
        final String createdTopic;

        private TopicCreatedCondition(final String topic) {
            createdTopic = topic;
        }

        @Override
        public boolean conditionMet() {
            return broker.kafkaServer().zkClient().getAllTopicsInCluster().contains(createdTopic) && broker.kafkaServer().metadataCache().contains(createdTopic);
        }
    }

    private static final class KafkaEmbedded {
        private static final Logger log = LoggerFactory.getLogger(KafkaEmbedded.class);
        private static final String DEFAULT_ZK_CONNECT = "127.0.0.1:2181";
        private final Properties effectiveConfig;
        private final KafkaServer kafkaServer;
        private final Path path;
        private final File logDir;

        public KafkaEmbedded(final Properties effectiveConfig) throws IOException {
            path = Files.createTempDirectory("KafkaEmbedded");
            logDir = path.toFile();
            this.effectiveConfig = effectiveConfigFrom(effectiveConfig);
            final KafkaConfig kafkaConfig = new KafkaConfig(this.effectiveConfig, true);
            log.debug("Starting embedded Kafka broker (with log.dirs={} and ZK ensemble at {}) ...", logDir, zookeeperConnect());
            kafkaServer = kafka.utils.TestUtils.createServer(kafkaConfig, Time.SYSTEM);
            log.debug("Startup of embedded Kafka broker at {} completed (with ZK ensemble at {}) ...", brokerList(), zookeeperConnect());
        }

        private Properties effectiveConfigFrom(final Properties initialConfig) {
            final Properties effectiveConfig = new Properties();
            effectiveConfig.put(KafkaConfig$.MODULE$.BrokerIdProp(), 0);
            effectiveConfig.put(KafkaConfig$.MODULE$.HostNameProp(), "127.0.0.1");
            effectiveConfig.put(KafkaConfig$.MODULE$.PortProp(), "9092");
            effectiveConfig.put(KafkaConfig$.MODULE$.NumPartitionsProp(), 1);
            effectiveConfig.put(KafkaConfig$.MODULE$.AutoCreateTopicsEnableProp(), true);
            effectiveConfig.put(KafkaConfig$.MODULE$.MessageMaxBytesProp(), 1000000);
            effectiveConfig.put(KafkaConfig$.MODULE$.ControlledShutdownEnableProp(), true);

            effectiveConfig.putAll(initialConfig);
            effectiveConfig.setProperty(KafkaConfig$.MODULE$.LogDirProp(), logDir.getAbsolutePath());
            return effectiveConfig;
        }

        public String zookeeperConnect() {
            return effectiveConfig.getProperty("zookeeper.connect", DEFAULT_ZK_CONNECT);
        }

        public String brokerList() {
            return String.join(":", kafkaServer.config().hostName(), Integer.toString(kafkaServer.boundPort(ListenerName.forSecurityProtocol(SecurityProtocol.PLAINTEXT))));
        }

        public void stop() throws IOException {
            log.debug("Shutting down embedded Kafka broker at {} (with ZK ensemble at {}) ...", brokerList(), zookeeperConnect());
            kafkaServer.shutdown();
            kafkaServer.awaitShutdown();
            log.debug("Removing temp logs.dir at {} ...", logDir);
            Files.walk(path)
                    .sorted(Comparator.reverseOrder())
                    .map(Path::toFile)
                    .forEach(File::deleteOnExit);
            log.debug("Shutdown of embedded Kafka broker at {} completed (with ZK ensemble at {}) ...", brokerList(), zookeeperConnect());
        }

        public void createTopic(final String topic,
                                final int partitions,
                                final short replication,
                                final Map<String, String> topicConfig) {
            log.debug("Creating topic { name: {}, partitions: {}, replication: {}, config: {} }", topic, partitions, replication, topicConfig);

            final Properties properties = new Properties();
            properties.put(AdminClientConfig.BOOTSTRAP_SERVERS_CONFIG, brokerList());

            try (final AdminClient adminClient = AdminClient.create(properties)) {
                final NewTopic newTopic = new NewTopic(topic, partitions, replication);
                newTopic.configs(topicConfig);
                adminClient.createTopics(Collections.singleton(newTopic)).all().get();
            } catch (InterruptedException | ExecutionException e) {
                e.printStackTrace();
            }
        }

        KafkaServer kafkaServer() {
            return kafkaServer;
        }
    }

    private static final class ZooKeeperEmbedded {

        private static final Logger log = LoggerFactory.getLogger(ZooKeeperEmbedded.class);
        private final TestingServer server;

        public ZooKeeperEmbedded() throws Exception {
            log.debug("Starting embedded ZooKeeper server...");
            this.server = new TestingServer(2181);
            log.debug("Embedded ZooKeeper server at {} uses the temp directory at {}", server.getConnectString(), server.getTempDirectory());
        }

        public void stop() throws IOException {
            log.debug("Shutting down embedded ZooKeeper server at {} ...", server.getConnectString());
            server.close();
            log.debug("Shutdown of embedded ZooKeeper server at {} completed", server.getConnectString());
        }

        public String connectString() {
            return server.getConnectString();
        }

        public String hostname() {
            return connectString().substring(0, connectString().lastIndexOf(':'));
        }
    }
}

